{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# FashionMNIST classification with Multilayer perceptrons\n",
    "\n",
    "https://github.com/zalandoresearch/fashion-mnist"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "import matplotlib.pyplot as plt\n",
    "import numpy as np\n",
    "from torchvision import datasets\n",
    "import torch\n",
    "import torch.nn as nn\n",
    "import torch.optim as optim"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# Preparing the dataset\n",
    "\n",
    "## Download and convert to float"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "train_data = datasets.FashionMNIST('data', download=True, train=True)\n",
    "# we need FloatTensors as input\n",
    "train_X = train_data.data.float()\n",
    "train_y = train_data.targets\n",
    "test_data = datasets.FashionMNIST('data', download=True, train=False)\n",
    "test_X = test_data.data.float()\n",
    "test_y = test_data.targets"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## IMPORTANT: Inspect the dataset manually\n",
    "\n",
    "We will plot a random sample with its labels:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "labels = ['T-shirt/top', 'Trouser', 'Pullover', 'Dress', 'Coat',\n",
    "'Sandal', 'Shirt', 'Sneaker', 'Bag', 'Ankle boot']\n",
    "\n",
    "idx = np.random.randint(len(train_X))\n",
    "sample_X = train_X[idx].numpy()\n",
    "sample_y = train_y[idx].numpy()\n",
    "print(\"Label: {}\".format(labels[sample_y]))\n",
    "plt.imshow(sample_X, 'gray')"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### Are the classes equally distributed?"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "np.unique(train_y.numpy(), return_counts=True)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Flatten images"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "print(\"Before flattening:\")\n",
    "print(\"Train size:\", train_X.size(), train_y.size())\n",
    "print(\"Test size:\", test_X.size(), test_y.size())\n",
    "train_X = train_X.view(-1, 28 * 28).squeeze(1)\n",
    "test_X = test_X.view(-1, 28 * 28).squeeze(1)\n",
    "print(\"\\nAfter flattening:\")\n",
    "print(\"Train size:\", train_X.size(), train_y.size())\n",
    "print(\"Test size:\", test_X.size(), test_y.size())"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Splitting the train set into train and dev set"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "all_idx = np.arange(len(train_X))\n",
    "train_idx = all_idx[:50000]\n",
    "dev_idx = all_idx[50000:]\n",
    "dev_X = train_X[dev_idx]\n",
    "dev_y = train_y[dev_idx]\n",
    "train_X = train_X[train_idx]\n",
    "train_y = train_y[train_idx]\n",
    "print(\"Train size:\", train_X.size(), train_y.size())\n",
    "print(\"Dev size:\", dev_X.size(), dev_y.size())\n",
    "print(\"Test size:\", test_X.size(), test_y.size())"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Helper class that creates minibatches"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "class BatchedIterator:\n",
    "    def __init__(self, X, y, batch_size):\n",
    "        self.X = X\n",
    "        self.y = y\n",
    "        self.batch_size = batch_size\n",
    "        self.batch_starts = list(range(0, len(self.X), batch_size))\n",
    "    \n",
    "    def iterate_once(self):\n",
    "        for start in range(0, len(self.X), self.batch_size):\n",
    "            end = start + self.batch_size\n",
    "            yield self.X[start:end], self.y[start:end]"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "train_iter = BatchedIterator(train_X, train_y, 33333)\n",
    "for batch in train_iter.iterate_once():\n",
    "    print(batch[0].size(), batch[1].size())"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# Neural network\n",
    "\n",
    "## Definition"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "class SimpleClassifier(nn.Module):\n",
    "    def __init__(self, input_dim, output_dim, hidden_dim):\n",
    "        super().__init__()\n",
    "        self.input_layer = nn.Linear(input_dim, hidden_dim)\n",
    "        self.relu = nn.ReLU()\n",
    "        self.output_layer = nn.Linear(hidden_dim, output_dim)\n",
    "        \n",
    "    def forward(self, X):\n",
    "        h = self.input_layer(X)\n",
    "        h = self.relu(h)\n",
    "        out = self.output_layer(h)\n",
    "        return out"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Model instance"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "model = SimpleClassifier(\n",
    "    input_dim=train_X.size(1),\n",
    "    output_dim=10,\n",
    "    hidden_dim=50\n",
    ")\n",
    "# we could move it to cuda at this point using model.cuda()\n",
    "model"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Loss function and optimizer"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "criterion = nn.CrossEntropyLoss()\n",
    "optimizer = optim.Adam(model.parameters())"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Sanity check\n",
    "\n",
    "The model should perform close to random at this point"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "test_pred = model(test_X).max(axis=1)[1]\n",
    "test_acc = torch.eq(test_pred, test_y).sum().float() / len(test_X)\n",
    "test_acc"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Training loop"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "batch_size = 1000\n",
    "train_iter = BatchedIterator(train_X, train_y, batch_size)\n",
    "dev_iter = BatchedIterator(dev_X, dev_y, batch_size)\n",
    "test_iter = BatchedIterator(test_X, test_y, batch_size)\n",
    "\n",
    "all_train_loss = []\n",
    "all_dev_loss = []\n",
    "all_train_acc = []\n",
    "all_dev_acc = []\n",
    "\n",
    "for epoch in range(10):\n",
    "    # training loop\n",
    "    for bi, (batch_x, batch_y) in enumerate(train_iter.iterate_once()):\n",
    "        y_out = model(batch_x)\n",
    "        loss = criterion(y_out, batch_y)\n",
    "        optimizer.zero_grad()\n",
    "        loss.backward()\n",
    "        optimizer.step()\n",
    "        \n",
    "    # one train epoch finished, evaluate on the train and the dev set (NOT the test)\n",
    "    train_out = model(train_X)\n",
    "    train_loss = criterion(train_out, train_y)\n",
    "    all_train_loss.append(train_loss.item())\n",
    "    train_pred = train_out.max(axis=1)[1]\n",
    "    train_acc = torch.eq(train_pred, train_y).sum().float() / len(train_X)\n",
    "    all_train_acc.append(train_acc)\n",
    "    \n",
    "    dev_out = model(dev_X)\n",
    "    dev_loss = criterion(dev_out, dev_y)\n",
    "    all_dev_loss.append(dev_loss.item())\n",
    "    dev_pred = dev_out.max(axis=1)[1]\n",
    "    dev_acc = torch.eq(dev_pred, dev_y).sum().float() / len(dev_X)\n",
    "    all_dev_acc.append(dev_acc)\n",
    "    \n",
    "    print(f\"Epoch: {epoch}\\n  train accuracy: {train_acc}  train loss: {train_loss}\")\n",
    "    print(f\"  dev accuracy: {dev_acc}  dev loss: {dev_loss}\")"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Test accuarcy"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "test_pred = model(test_X).max(axis=1)[1]\n",
    "test_acc = torch.eq(test_pred, test_y).sum().float() / len(test_X)\n",
    "test_acc"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Plot loss functions, accuracies"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "plt.plot(all_train_loss, label='train')\n",
    "plt.plot(all_dev_loss, label='dev')\n",
    "plt.legend()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": [
    "plt.plot(all_train_acc, label='train')\n",
    "plt.plot(all_dev_acc, label='dev')\n",
    "plt.legend()"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# Exercises\n",
    "\n",
    "1. Try changing the model and training parameters (hidden size, epoch, batch size).\n",
    "1. Add another hidden layer to the model.\n",
    "1. All input values are between 0 to 255 (RGB brightness values). Normalize them (scale them to between 0 and 1). You should see a small increase in accuracy.\n",
    "1. Try standardizing the input values (mean 0, std 1).\n",
    "1. Implement early stopping. Instead of a fix number of epochs, train until the model stops improving on the dev set. If the dev loss stop decreasing for a few epochs, training should stop.\n",
    "1. ADVANCED: replace the network with a convolutional neural network. Convolutions should work on 2D images, change the preprocessing steps accordingly. You should see a big increase in accuracy."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": []
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": []
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.8.5"
  },
  "toc": {
   "base_numbering": 1,
   "nav_menu": {
    "height": "339.633px",
    "width": "252px"
   },
   "number_sections": true,
   "sideBar": true,
   "skip_h1_title": false,
   "title_cell": "Table of Contents",
   "title_sidebar": "Contents",
   "toc_cell": false,
   "toc_position": {},
   "toc_section_display": "block",
   "toc_window_display": true
  }
 },
 "nbformat": 4,
 "nbformat_minor": 4
}
